from mlhandmade.kernel_methods.svm import (
    SupportVectorClassifier,
    SupportVectorRegressor
)

from mlhandmade.kernel_methods.kernel_ridge import KernelRidge

__all__ = [
    "SupportVectorClassifier",
    "SupportVectorRegressor",
    "KernelRidge"
]


from mlhandmade.ensemble.forest import (
    RandomForestClassifier,
    RandomForestRegressor
)

from mlhandmade.ensemble.adaboost import (
    AdaBoostClassifier,
    AdaBoostRegressor
)

__all__ = [
    "RandomForestClassifier",
    "RandomForestRegressor",
    "AdaBoostClassifier",
    "AdaBoostRegressor"
]


from abc import abstractmethod
import numpy as np
from scipy.special import xlogy

from ..base import BaseEstimator
from ..tree import DecisionTreeClassifier, DecisionTreeRegressor
from ..utils.validations import check_random_state, _num_samples

class BaseAdaBoost(BaseEstimator):
    """
    Base class for AdaBoost algorithms
    base on decision trees

    Parameters
    ----------
    base_tree : BaseDecisionTree class
        DecisionTreeClassifier in case of classification
        DecisionTreeRegressor in case of regression
    estimator_kwargs : dict
        keyword arguments for estimator
    n_estimators : int (default: 50)
        number of estimators
    learning_rate : float (default: 1.0)
        hyperparameter as a `n_estimators`. `learning_rate`
        controls contribution of each estimator in ensemble
    """
    def __init__(
        self,
        base_tree,
        estimator_kwargs,
        n_estimators=50,
        learning_rate=1.0,
        random_state=0
    ):
        self.base_tree = base_tree
        self.n_estimators = n_estimators
        self.estimator_kwargs = estimator_kwargs
        self.learning_rate = learning_rate
        self.rgen = check_random_state(random_state)

    def _fit(self, X, y):
        self.estimators_ = []
        self.estimator_weights_ = np.zeros(self.n_estimators, dtype=np.float64)
        self.estimator_errors_ = np.ones(self.n_estimators, dtype=np.float64)
        sample_weight = np.ones(self.n_samples) / self.n_samples

        for iboost in range(self.n_estimators):
            # get different random states for each estimator to avoid
            # choice of the same feature due to `max_feature`
            random_seed = self.rgen.randint(1000)
            sample_weight, estimator_weight, estimator_error = self._boost(
                iboost, X, y, sample_weight, random_state=check_random_state(random_seed)
            )

            if sample_weight is None: # case of bad fitting
                break
            self.estimator_weights_[iboost] = estimator_weight
            self.estimator_errors_[iboost] = estimator_error

            if estimator_error == 0: # case of perfect fit
                break

            sample_weight_sum = np.sum(sample_weight)

            if sample_weight_sum <= 0: # sum sample_weight in tree should be positive
                break

            if iboost < self.n_estimators - 1:
                sample_weight /= sample_weight_sum

    @abstractmethod
    def _boost(self, iboost, X, y, sample_weight, random_state):
        raise NotImplementedError()

def _samme_proba(estimator, n_classes, X):
    """
    Get class probabilities in case of "SAMME.R"
    """
    proba = estimator.predict_proba(X)
    np.clip(proba, np.finfo(proba.dtype).eps, None, out=proba)
    log_proba = np.log(proba)

    return (n_classes - 1) * (
        log_proba - (1.0 / n_classes) * log_proba.sum(axis=1)[:, np.newaxis]
    )

class AdaBoostClassifier(BaseAdaBoost):
    """
    AdaBoost multiclass classifier with SAMME algorithm

    Parameters
    ----------
    n_estimators : int (default: 10)
        number of trees
    criterion : ["gini", "entropy"] (default: "gini")
        criterion for DecisionTreeClassifier
    max_depth : int or inf (default: 1)
        max depth of trees, default value is 1, i.e.
        decision stumps
    min_samples_leaf : int (default: 1)
        min samples in leaf to split node
    max_features : "sqrt", int or None (default: "sqrt")
        features to consider on each split
    learning_rate : float (default: 1.0)
        hyperparameter as a `n_estimators`. `learning_rate`
        controls contribution of each estimator in ensemble
    algorithm : ["SAMME", "SAMME.R"] (default: SAMME.R)
        "SAMME" algorithm uses discrete class labels,
        "SAMME.R" uses their probabilities instead,
        so "SAMME.R" is more efficient
    """
    def __init__(
        self,
        n_estimators=50,
        *,
        criterion="gini",
        max_depth=1,
        min_samples_leaf=1,
        max_features="sqrt",
        learning_rate=1.0,
        algorithm="SAMME.R",
        random_state=0
    ):
        if algorithm not in ("SAMME", "SAMME.R"):
            raise ValueError(
                "Algorithm must be \"SAMME\" or \"SAMME.R\","
                f" got {algorithm!r} instead."
            )
        
        self.algorithm = algorithm

        estimator_kwargs={
            "criterion": criterion,
            "max_depth": max_depth,
            "min_samples_leaf": min_samples_leaf,
            "max_features": max_features
        }
        super().__init__(
            base_tree=DecisionTreeClassifier,
            estimator_kwargs=estimator_kwargs,
            n_estimators=n_estimators,
            learning_rate=learning_rate,
            random_state=random_state
        )

    def _fit(self, X, y):
        self.classes_ = np.unique(y)
        self.n_classes_ = self.classes_.shape[0]
        return super()._fit(X, y)

    def _boost(self, iboost, X, y, sample_weight, random_state):
        self.estimator_kwargs["random_state"] = random_state
        if self.algorithm == "SAMME.R":
            return self._boost_real(iboost, X, y, sample_weight)
        elif self.algorithm == "SAMME":
            return self._boost_discrete(iboost, X, y, sample_weight)
    
    def _boost_real(self, iboost, X, y, sample_weight):
        tree = self.base_tree(**self.estimator_kwargs)
        tree.fit(X, y, sample_weight=sample_weight)
        self.estimators_.append(tree)

        y_predict_proba = tree.predict_proba(X)

        y_predict = self.classes_.take(np.argmax(y_predict_proba, axis=1), axis=0)

        incorrect = y_predict != y

        estimator_error = np.mean(np.average(incorrect, weights=sample_weight, axis=0))
        if estimator_error <= 0:
            return sample_weight, 1.0, 0.0

        classes = self.classes_
        n_classes = self.n_classes_
        y_codes = np.array([-1.0 / (n_classes - 1), 1.0])
        y_coding = y_codes.take(classes == y[:, np.newaxis])

        proba = y_predict_proba
        np.clip(proba, np.finfo(proba.dtype).eps, None, out=proba)

        estimator_weight = (
            -1.0
            * self.learning_rate
            * ((n_classes - 1.0) / n_classes)
            * xlogy(y_coding, y_predict_proba).sum(axis=1)
        )

        if iboost != self.n_estimators - 1:
            sample_weight *= np.exp(
                estimator_weight * ((sample_weight > 0) | (estimator_weight < 0))
            )
        
        return sample_weight, 1.0, estimator_error

    def _boost_discrete(self, iboost, X, y, sample_weight):
        tree = self.base_tree(**self.estimator_kwargs)
        tree.fit(X, y, sample_weight=sample_weight)
        self.estimators_.append(tree)

        y_predict = tree.predict(X)

        incorrect = y_predict != y

        estimator_error = np.mean(np.average(incorrect, weights=sample_weight, axis=0))

        if estimator_error <= 0:
            return sample_weight, 1.0, 0.0

        n_classes = self.n_classes_

        if estimator_error >= 1.0 - (1.0 / n_classes):
            self.estimators_.pop(-1)
            if len(self.estimators_) == 0:
                raise ValueError(
                    "AdaBoostClassifier is worse than random,"
                    " ensemble can not be fit."
                )
            return None, None, None

        estimator_weight = self.learning_rate * (
            np.log((1.0 - estimator_error) / estimator_error) + np.log(n_classes - 1.0)
        )

        if iboost != self.n_estimators - 1:
            sample_weight = np.exp(
                np.log(sample_weight)
                + estimator_weight * incorrect * (sample_weight > 0)
            )

        return sample_weight, estimator_weight, estimator_error

    def decision_function(self, X):
        n_classes = self.n_classes_
        classes = self.classes_[:, np.newaxis]

        if self.algorithm == "SAMME.R":
            pred = sum(
                _samme_proba(estimator, n_classes, X) for estimator in self.estimators_
            )
        elif self.algorithm == "SAMME":
            pred = sum(
                (estimator.predict(X) == classes).T * w
                for estimator, w in zip(self.estimators_, self.estimator_weights_)
            )

        pred /= self.estimator_weights_.sum()
        return pred

    def _predict(self, X):
        pred = self.decision_function(X)
        
        return self.classes_.take(np.argmax(pred, axis=1), axis=0)

class AdaBoostRegressor(BaseAdaBoost):
    """
    AdaBoost regressor with AdaBoost.R2 algorithm

    Parameters
    ----------
    n_estimators : int (default: 10)
        number of trees
    criterion : ["mse", "mae"] (default: "mse")
        criterion for DecisionTreeClassifier
    max_depth : int or inf (default: 3)
        max depth of trees
    min_samples_leaf : int (default: 1)
        min samples in leaf to split node
    max_features : "sqrt", int or None (default: None)
        features to consider on each split
    loss : ["linear", "square", "exponential"] (default: "linear")
        loss function to use when updating the weights after each
        boosting iteration.
    """
    def __init__(
        self,
        n_estimators=50,
        *,
        criterion="mse",
        max_depth=3,
        min_samples_leaf=1,
        max_features=None,
        learning_rate=1.0,
        loss="linear",
        random_state=0,
    ):
        if loss not in ("linear", "square", "exponential"):
            raise ValueError(
                    "Loss must be \"linear\", \"square\" or \"exponential\","
                    f" got {loss!r} instead."
                )
        
        self.loss = loss
        estimator_kwargs={
            "criterion": criterion,
            "max_depth": max_depth,
            "min_samples_leaf": min_samples_leaf,
            "max_features": max_features
        }
        super().__init__(
            base_tree=DecisionTreeRegressor,
            estimator_kwargs=estimator_kwargs,
            n_estimators=n_estimators,
            learning_rate=learning_rate,
            random_state=random_state
        )

    def _fit(self, X, y):
        return super()._fit(X, y)

    def _boost(self, iboost, X, y, sample_weight, random_state):
        self.estimator_kwargs["random_state"] = random_state
        tree = self.base_tree(**self.estimator_kwargs)
        bootstrap_idx = random_state.choice(
            np.arange(self.n_samples),
            size=self.n_samples,
            replace=True,
            p=sample_weight,
        )
        tree.fit(X[bootstrap_idx], y[bootstrap_idx])
        self.estimators_.append(tree)
        y_predict = tree.predict(X)

        error_vect = np.abs(y_predict - y)
        sample_mask = sample_weight > 0
        masked_sample_weight = sample_weight[sample_mask]
        masked_error_vect = error_vect[sample_mask]

        error_max = masked_error_vect.max()
        if error_max != 0:
            masked_error_vect /= error_max

        if self.loss == "square":
            masked_error_vect **= 2
        elif self.loss == "exponential":
            masked_error_vect = 1.0 - np.exp(-masked_error_vect)

        estimator_error = (masked_sample_weight * masked_error_vect).sum()

        if estimator_error <= 0:
            return sample_weight, 1.0, 0.0
        elif estimator_error >= 0.5:
            if len(self.estimators_) > 1:
                self.estimators_.pop(-1)
            return None, None, None
        
        beta = estimator_error / (1.0 - estimator_error)
        estimator_weight = self.learning_rate * np.log(1.0 / beta)

        if iboost != self.n_estimators - 1:
            sample_weight[sample_mask] *= np.power(
                beta, (1.0 - masked_error_vect) * self.learning_rate
            )
        
        return sample_weight, estimator_weight, estimator_error

    def _get_median_predict(self, X, limit):
        prediction = np.array([e.predict(X) for e in self.estimators_[:limit]]).T
        sorted_idx = np.argsort(prediction, axis=1)
        weight_cdf = np.cumsum(self.estimator_weights_[sorted_idx], axis=1)
        median_or_above = weight_cdf >= 0.5 * weight_cdf[:, -1][:, np.newaxis]
        median_idx = median_or_above.argmax(axis=1)

        median_estimators = sorted_idx[np.arange(_num_samples(X)), median_idx]

        return prediction[np.arange(_num_samples(X)), median_estimators]

    def _predict(self, X):
        return self._get_median_predict(X, len(self.estimators_))


from abc import abstractmethod
import numpy as np
from ..preprocessing import data_shuffle
from ..utils.validations import check_random_state

class BaseOptimizer:
    """
    Base class for optimizers of finite sums.
    """
    @abstractmethod
    def update(self, grad, X, y, w):
        raise NotImplementedError()

class GD(BaseOptimizer):
    """
    Gradient descent optimizer:

    w_new = w - eta * mean(grad(L(X, y, w))),
    L - loss
    X - full dataset samples
    y - full dataset targets
    w - weights

    Parameters
    ----------
    eta : float
        learning rate
        
    Returns
    -------
    w : ndarray of shape(D+1,)
        weights
    """
    def __init__(self, eta):
        self.eta = eta

    def update(self, grad, X, y, w):
        w -= self.eta * np.mean(grad(X, y, w), axis=0)
        return w

class SGD(BaseOptimizer):
    """
    Stochastic gradient descent optimizer:

    w_new = w - grad(L(xi, yi, w))
    L - loss
    xi - one sample from dataset
    yi - one target from dataset
    w - weights

    Parameters
    ----------
    eta : float
        learning rate
    random_state : int, None, np.random.RandomState (default: 0)
        random seed to reproduce calculations
    shuffle : bool (default: True)
        shuffle data before each iteration if True
        
    Returns
    -------
    w : ndarray of shape(D+1,)
        weights
    """
    def __init__(self, eta, random_state=0, shuffle=True):
        self.eta = eta
        self.rgen = check_random_state(random_state)
        self.shuffle = shuffle

    def update(self, grad, X, y, w):
        if self.shuffle:
            X, y = data_shuffle(X, y, random_state=self.rgen)
        for xi, target in zip(X, y):
            w -= self.eta * grad(xi, target, w)
        return w

class BatchGD(BaseOptimizer):
    """
    Batch gradient descent optimizer

    w_new = w - mean(grad(L(Xs, ys, w))),
    L - loss
    Xs - random subsample from X (batch)
    ys - random subsample from y (batch)
    w - weights

    Parameters
    ----------
    eta : float
        learning rate
    batch_size : int in (0, N]
        size of a batch
    random_state : int, None, np.random.RandomState (default: 0)
        random seed to reproduce calculations
    shuffle : bool (default: True)
        shuffle data before each iteration if True

    Returns
    -------
    w : ndarray of shape(D+1,)
        weights

    SGD is a special case of BatchGD with `batch_size=1`
    """
    def __init__(self, eta, batch_size, random_state=0, shuffle=True):
        self.eta = eta
        self.batch_size = batch_size
        self.rgen = check_random_state(random_state)
        self.shuffle = shuffle

    def update(self, grad, X, y, w):
        for xb, tb in self._batch_iterator(X, y):
            w -= self.eta * np.mean(grad(xb, tb, w), axis=0)
        return w

    def _batch_iterator(self, X, y):
        if self.batch_size > X.shape[0] or self.batch_size <= 0:
            raise ValueError(
                    "batch_size must be in (0, num_sampes],"
                    f" got {self.batch_size}."
                )
        
        idx = np.arange(X.shape[0])
        if self.shuffle:
            self.rgen.shuffle(idx)

        for i in range(0, X.shape[0], self.batch_size):
            begin, end = i, min(i + self.batch_size, X.shape[0])
            yield X[idx[begin:end]], y[idx[begin:end]]

class SAG(BaseOptimizer):
    """
    Stochastic average gradient optimizer.
    This algorithm calculates gradients only on first
    iteration, then it updates only one gradient component
    on each iteration. In other its similar to GD

    Parameters
    ----------
    eta : float
        learning rate
    random_state : int, None, np.random.RandomState (default: 0)
        random seed to reproduce calculations
        
    Returns
    -------
    w : ndarray of shape(D+1,)
        weights
    """
    def __init__(self, eta, random_state=0):
        self.eta = eta
        self.rgen = check_random_state(random_state)

    @staticmethod
    def _setup_grads(grad, X, y, w):
        grads = grad(X, y, w)
        return grads

    def update(self, grad, X, y, w):
        if not hasattr(self, "grads_"):
            self.grads_ = self._setup_grads(grad, X, y, w)
        k = self.rgen.randint(0, X.shape[0])
        self.grads_[k] = grad(X[k], y[k], w)
        w -= self.eta * np.mean(self.grads_, axis=0)
        return w


from setuptools import setup, find_packages
from codecs import open
from os import path

__version__ = "0.1.0"

here = path.abspath(path.dirname(__file__))

with open(path.join(here, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

with open(path.join(here, "requirements.txt"), encoding="utf-8") as f:
    all_reqs = f.read().split("\n")

install_requires = [req.strip() for req in all_reqs if "git+" not in req]
dependency_links = [req.strip().replace("git+", "") for req in all_reqs if req.startswith("git+")]

setup(
    name="mlhandmade",
    version=__version__,
    description="Implemented some ML routines",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="gittasche",
    python_reqires=">=3.9",
    install_requires=install_requires,
    setup_requires=["numpy>=1.22.3", "scipy>=1.8.0", "matplotlib>=3.5.1"],
    dependency_links=dependency_links,
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False
)


import numpy as np
import numbers
from scipy.spatial.distance import cdist

from ..base import BaseEstimator
from ..utils.validations import check_random_state

class KMeans(BaseEstimator):
    """
    Lloyd k-means clustering algorirthm

    Parameters
    ----------
    n_clusters : int
        number of clusters
    init : ["random", "k-means++"] (default: "k-means++")
        way to initialize centroids
    max_iter : int (default: 300)
        number of iterations in Lloyd algorithm
    tol : float (default: 1e-4)
        tolerance value to check approximate convergence
    """
    def __init__(self, n_clusters, init="k-means++", max_iter=300, tol=1e-4, random_state=0):
        self.n_clusters = n_clusters
        self.init = init
        self.max_iter = max_iter
        self.tol = tol
        self.rgen = check_random_state(random_state)

        self.y_required = False

    def _validate_input(self):
        message = (
            "n_clusters must be int > 0,"
            f" got {self.n_clusters} of type {type(self.n_clusters).__name__}."
        )
        if not isinstance(self.n_clusters, numbers.Integral):
            raise ValueError(message)
        if not self.n_clusters > 0:
            raise ValueError(message)

        if self.init not in ["random", "k-means++"]:
            raise ValueError(
                "init must be \"random\" or \"k-means++\","
                f" got \"{self.init}\"."
            )

    def _fit(self, X, y=None):
        self._validate_input()
        X_mean = X.mean(axis=0)
        X -= X_mean

        centers_init = self._init_centroids(X)
        self.labels_, self.centers_ = self._lloyd_kmeans(X, centers_init)

    def _init_centroids(self, X):
        if self.init == "k-means++":
            centers, _ = self._kmeansplusplus(X)
        elif self.init == "random":
            seeds = self.rgen.permutation(self.n_samples)[:self.n_clusters]
            centers = X[seeds]
        
        return centers

    def _kmeansplusplus(self, X):
        centers = np.empty((self.n_clusters, self.n_features), dtype=X.dtype)

        center_id = self.rgen.randint(self.n_samples)
        indices = np.full(self.n_clusters, -1, dtype=int)
        centers[0] = X[center_id]
        indices[0] = center_id

        n_local_trials = 2 + int(np.log(self.n_clusters))
        closest_dist = cdist(centers[0, np.newaxis], X, metric="euclidean")
        current_pot = closest_dist.sum()

        for c in range(1, self.n_clusters):
            rand_vals = self.rgen.uniform(size=n_local_trials) * current_pot
            candidate_ids = np.searchsorted(np.cumsum(closest_dist), rand_vals)
            # ensure max(candidate_idx) > len(closest_dist)
            np.clip(candidate_ids, None, closest_dist.size - 1, out=candidate_ids)

            distances_to_candidates = cdist(X[candidate_ids], X, metric="euclidean")

            np.minimum(closest_dist, distances_to_candidates, out=distances_to_candidates)
            candidates_pot = distances_to_candidates.sum(axis=1)

            best_candidate = np.argmin(candidates_pot)
            current_pot = candidates_pot[best_candidate]
            closest_dist = distances_to_candidates[best_candidate]
            best_candidate = candidate_ids[best_candidate]

            centers[c] = X[best_candidate]
            indices[c] = best_candidate
        
        return centers, indices

    def _lloyd_kmeans(self, X, centers_init):
        centers = centers_init
        labels = np.full(self.n_samples, -1, dtype=np.int32)

        for _ in range(self.max_iter):
            distances = cdist(centers, X, metric="euclidean")
            labels_new = np.argmin(distances, axis=0)
            centers_new = np.array([
                np.mean(X[labels_new == cl], axis=0) for cl in range(self.n_clusters)
            ])

            # check strict convergence
            if np.array_equal(labels, labels_new):
                break
            # check approximate convergence
            elif np.allclose(centers, centers_new, atol=self.tol):
                break

            centers = centers_new
            # assign using slice to avoid dtype change
            labels[:] = labels_new

        return labels, centers

    def _predict(self, X):
        distances = cdist(self.centers_, X)
        return np.argmin(distances, axis=0)


import numpy as np

def check_pairwise(X, Y):
    """
    Ensure arrays X and Y to 
    calculate Gram matrix
    """
    X = np.atleast_2d(X)
    Y = np.atleast_2d(Y)

    if X.shape[1] != Y.shape[1]:
        raise ValueError(
            "X and Y must have equal last dimension,"
            f" but {X.shape[1]} != {Y.shape[1]}"
        )

    return X, Y

class LinearKernel:
    """
    Linear kernel: K = X x Y^T + c

    Parameters
    ----------
    coef0 : float (default: 0.0)
        bias value
    """
    def __init__(self, coef0=0.0):
        self.coef0 = coef0

    def __call__(self, X, Y):
        X, Y = check_pairwise(X, Y)
        return X @ Y.T + self.coef0

class PolyKernel:
    """
    Polynomial kernel: K = (gamma * X x Y^T + c)^d

    Parameters
    ----------
    coef0 : float (default: 0.0)
        bias value
    degree : int (default: 3)
        degree of polynomial
    gamma : float (default: None)
        coefficient, by default `gamma = 1 / X.shape[1]`
    """
    def __init__(self, coef0=0.0, degree=3, gamma=None):
        self.coef0 = coef0
        self.degree = degree
        self.gamma = gamma

    def __call__(self, X, Y):
        X, Y = check_pairwise(X, Y)
        if self.gamma is None:
            self.gamma = 1.0 / X.shape[1]
        return np.power(self.gamma * X @ Y.T + self.coef0, self.degree)

class RBFKernel:
    """
    RBF kernel: K = exp(-gamma * ||X - Y||^2)

    Parameters
    ----------
    gamma : float (default: None)
        coefficient, by default `gamma = 1 / X.shape[1]`
    """
    def __init__(self, gamma=None):
        self.gamma = gamma

    def __call__(self, X, Y):
        X, Y = check_pairwise(X, Y)
        if self.gamma is None:
            self.gamma = 1.0 / X.shape[1]

        K = self._euclidean_sq_distances(X, Y)
        K *= -self.gamma
        np.exp(K, out=K)
        return K

    @staticmethod
    def _euclidean_sq_distances(X, Y):
        # einsum("ij,ij->i", X, X) - squared row norms of X
        XX = np.einsum("ij,ij->i", X, X)[:, np.newaxis]
        YY = np.einsum("ij,ij->i", Y, Y)[np.newaxis, :]

        distances = XX + YY - 2 * X @ Y.T
        np.maximum(distances, 0, out=distances)

        # prevent floating point rounding error
        if X is Y:
            np.fill_diagonal(distances, 0)

        return distances